import numpy as np, tensorflow as tf, warnings, os
from tensorflow.keras.layers import (LSTM, BatchNormalization, Dense, Input, LeakyReLU)
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Adam, SGD
from tensorflow.keras.regularizers import l2

warnings.filterwarnings('ignore')  # remove TF clutter
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # or any {'0', '1', '2'}


def train_lstm(x, y, d_params):
    """ Creates a model and fits it with the given data and configuration."""

    h_params = {**d_params["h_params"], **d_params["timeseries"]["algo"]["lstm"]}

    layers = [Input((x.shape[1:]), name='input_lstm_0')]
    layers += [model_lstm_layers(h_params, layers[0])]
    layers += [model_final_layer(layers[-1], y, h_params)]
    model = Model(inputs=layers[0], outputs=layers[-1])

    optimizer = {
        'adam': Adam(learning_rate=h_params['learning_rate']),
        'sgd': SGD(learning_rate=h_params['learning_rate']),
    }

    model.compile(
        optimizer=optimizer[h_params['optimizer']],
        loss=d_params['metrics'][0], 
        metrics=d_params['metrics']
    )

    callbacks = [tf.keras.callbacks.ModelCheckpoint(f'/mnt/models/{d_params["mid"]}/{d_params["hid"]}.h5', save_best_only=True, save_freq='epoch', verbose=1)]
    callbacks += [
        tf.keras.callbacks.RemoteMonitor(
            root=d_params["url"],
            path=f'/api/v1/add-learn-log/{d_params["mid"]}/{d_params["hid"]}',
            send_as_json=True
        )
    ]

    # y = y.astype('float32')

    history = model.fit(x=x, y=y, batch_size=h_params['batch_size'], epochs=h_params['epochs'],
                             callbacks=callbacks, validation_split=0.3)

    return history, model


def model_lstm_layers(h_params, input_layer_0):

    flag_lstm_layer = False

    for i in range(h_params["LSTM_layers"]):
        return_sequences = True if h_params.get(
            'use_attention') else i < (h_params['LSTM_layers']-1)

        if i > 0:
            flag_lstm_layer = True
        act_to_use = h_params['activation'] if i != 0 else h_params['first_activation']

        # and not lag_lstm_layer:
        if h_params['LSTM_batch_norm'] and act_to_use not in ['sigmoid', 'tanh']:
            lstm_layer = BatchNormalization()(lstm_layer if flag_lstm_layer else input_layer_0)

        if act_to_use != "lrelu":
            if h_params['l1_l2'] != 0:
                lstm_layer = LSTM(units=h_params['LSTM_nodes'], activation=act_to_use, return_sequences=return_sequences, kernel_regularizer=l2(h_params['l1_l2']), recurrent_regularizer=l2(h_params['l1_l2']), bias_regularizer=l2(h_params['l1_l2'])
                                  )(lstm_layer if flag_lstm_layer else input_layer_0)
            else:
                lstm_layer = LSTM(units=h_params['LSTM_nodes'], activation=act_to_use, return_sequences=return_sequences)(
                    lstm_layer if flag_lstm_layer else input_layer_0)
        elif h_params['l1_l2'] != 0:
            lstm_layer = LSTM(units=h_params['LSTM_nodes'],
                              return_sequences=return_sequences,
                              kernel_regularizer=l2(0.01), recurrent_regularizer=l2(0.01), bias_regularizer=l2(0.01)
                              )(lstm_layer if flag_lstm_layer else input_layer_0)
        else:
            lstm_layer = LSTM(units=h_params['LSTM_nodes'], return_sequences=return_sequences)(
                lstm_layer if flag_lstm_layer else input_layer_0)

        if act_to_use == 'lrelu':
            lstm_layer = LeakyReLU()(lstm_layer)

        if h_params['LSTM_batch_norm'] and act_to_use in ['sigmoid', 'tanh']:
            lstm_layer = BatchNormalization()(lstm_layer)

    if h_params['LSTM_batch_norm'] and act_to_use not in ['sigmoid', 'tanh']:
        lstm_layer = BatchNormalization()(lstm_layer)

    return lstm_layer


def model_final_layer(final_layer, y, h_params):


    if h_params['final_activation'] != 'lrelu':
        final_layer = Dense(
            y.shape[-1], activation=h_params['final_activation'])(final_layer)
    else:
        final_layer = Dense(y.shape[-1])(final_layer)
        final_layer = LeakyReLU()(final_layer)
    return final_layer



import pandas as pd, numpy as np, datetime
from dateutil import relativedelta
from sklearn.preprocessing import MinMaxScaler


pd.options.mode.chained_assignment = None
from vai_utils import data


DEP = "count_tx7zf4t"  # count parameter for statistical methods


def main(dataset, d_params):
    '''
    Preprocess the dataset
    '''

    # Assume count is the feature if none specified
    if not d_params["x"]:  # TODO support add features for reg input too.
        dataset[DEP] = 1  # count instances
        dataset = add_counts(dataset, d_params)

        # Adding Features to the dataset TODO check
        dataset = add_features(dataset, d_params).copy()
        
        # TODO find d_types for validating them
        x_columns = dataset.columns.drop(DEP)
        d_params["x"] = {k: {"type": "NUMBER"} for k in x_columns}
        d_params["y"] = {DEP: {"type": "NUMBER"}}

        x, y, data_mappings = data.validate(d_params, dataset)  # Bin Data and Format NN Input
        dataset = pd.concat([x, y], axis=1)
    else:
        raise Exception("Error: Currently only support 'timeseries' with 'datetime'")
    
    # Reshape 2d to 3d timeseries
    x_seq, y_seq = split_x_y_sequences_in_out(
        np.concatenate((x, y), axis=1),
        n_steps_in=d_params["timeseries"]["sort"]["datetime"]["steps_in"], n_steps_out=d_params["timeseries"]["sort"]["datetime"]["steps_out"]
    )

    return x_seq.copy(), y_seq.copy(), data_mappings



def add_counts(dataset, d_params):
    """Add Statistics, but only works for "x" not definted"""

    if d_params["timeseries"]["sort"]["datetime"]["step"] in ["hour", "day"] or "T" in d_params["timeseries"]["sort"]["datetime"]["step"]:

        # Resample the dataframe by hour and sum the count_tx7zf4t values
        dataset = dataset.sort_values(by='datetime')

        # TODO add preiod (e.g. 3T) between start and time
        if d_params["timeseries"]["sort"]["datetime"]['step'] == 'hour':
            return dataset.resample('H').sum().copy()
        elif d_params["timeseries"]["sort"]["datetime"]['step'] == 'day':
            return dataset.resample('D').sum().copy()
        elif 'T' in d_params["timeseries"]["sort"]["datetime"]['step'] == d_params["timeseries"]['step']:
            return dataset.resample(d_params["timeseries"]["sort"]["datetime"]['step']).sum().copy()
        else:
            raise Exception(f"Error: 'datetime' only compatible with 'step'")
    else:
        raise Exception(f"Error: Currently only support datetime sorting")


def add_features(df, d_params):
    '''Add features to the dataframe
    - dow: day of week
    - doy: day of year
    - hour: hour of day
    - month: month of year
    - nld: non labour day
    - ct_lag_{i}: lagged value of ct
    - ct_delta_{i}: delta of ct
    '''
    df['dow'] = df.index.dayofweek
    df['hour'] = df.index.hour
    df['doy'] = df.index.dayofyear
    df['month'] = df.index.month
    df['day'] = df.index.day

    # Hours that the bussiness is not working TODO update this
    if "time-window" in d_params['timeseries']:
        start_time, end_time = d_params["timeseries"]["time-window"]
        df = df.between_time(start_time, end_time)

    # Calc Lag and Delta
    for i in range(1, d_params["timeseries"]["sort"]["datetime"]['lag']):
        df.loc[:, f'ct_lag_{i}'] = df[DEP].shift(i)

    for i in range(1, d_params["timeseries"]["sort"]["datetime"]['delta']):
        df.loc[:, f'ct_delta_{i}'] = df[DEP].diff(i)

    # Non Labour Days
    if d_params["timeseries"]["sort"]["datetime"]['add_nld']:
        df = add_non_labour_days(df)

    # MeanStd hour
    if d_params["timeseries"]["sort"]["datetime"]['add_meanstd']:
        df = add_mean_std(df, 'hour', DEP)
        df = add_mean_std(df, 'dow', DEP)
        df = add_mean_std(df, 'day', DEP)

    # Sort the columns
    df = df[sorted(df.columns)]
    
    return df.dropna()


def fix_na(df):
    for c in df.columns:
        if df[c].dtype != 'object':
            df[c] = df[c].replace([np.inf, -np.inf], np.NaN)
            df[c] = df[c].interpolate(method='linear', limit_direction='both')
    return df


def add_non_labour_days(df):
    '''Add a column with 1 if the day is a non labour day and 0 otherwise'''
    nld = get_non_labour_days()
    dfx = df.copy()
    dfx['month-day'] = [(m, d) for m, d in zip(df.index.month, df.index.day)]
    df['nld'] = (dfx['month-day'].apply(lambda x: x in nld)).astype(int)
    return df


def get_non_labour_days():
    '''Harcoded for 2022'''
    return [
        (1, 1), (1, 17), (2, 21),
        (5, 30), (7, 4), (9, 5),
        (10, 10), (11, 11), (11, 24),
        (12, 26)
    ]


def day_init_end_daylight_adj(year=2022):
    # Get the date for the start of daylight saving time
    start_date = datetime.datetime(year, 3, 1)  # Start from the 1st of March

    # Find the second Sunday in March
    while start_date.weekday() != 6:  # Check if the current day is not Sunday
        # Move to the next day
        start_date += relativedelta.relativedelta(days=1)

    # Get the date for the end of daylight saving time
    end_date = datetime.datetime(year, 11, 1)  # Start from the 1st of November

    # Find the first Sunday in November
    while end_date.weekday() != 6:  # Check if the current day is not Sunday
        end_date += relativedelta.relativedelta(days=1)  # Move to the next day

    # Get the day of the year when daylight saving time ends
    return start_date, end_date


def adjust_daylight_saving_time(df):
    start_date, end_date = day_init_end_daylight_adj()
    df['dst'] = (
        df.index >= start_date) & (df.index <= end_date)
    # df[df['daylight_saving_time']].index = df[df['daylight_saving_time']#].index - relativedelta.relativedelta(hours=1)
    df['dst'] = df['dst'].astype(int)
    return df


def add_mean_std(df, field, dep):
    df = df.copy().reset_index(drop=False)
    df0 = pd.DataFrame(df.groupby(field)[dep].mean()).reset_index(
        drop=False).rename(columns={dep: f'{field}_mean'})
    df1 = pd.DataFrame(df.groupby(field)[DEP].std()).reset_index(
        drop=False).rename(columns={dep: f'{field}_std'})
    dfx = pd.merge(df, df0, on=field, how='left')
    dfx = pd.merge(dfx, df1, on=field, how='left')
    dfx = dfx.set_index('datetime', drop=True)
    return dfx


def split_x_y_sequences_in_out(sequences, n_steps_in: int = 3, n_steps_out: int = 1):
    """Split a multivariate sequence into X,y samples

    Args: 
        sequences (numpy array)
        n_steps_in (int): number of input time steps
        n_steps_out (int): number of output time steps

    Returns:
        X (numpy array): input sequences
        y (numpy array): output sequences
    """
    X = [sequences[i:i+n_steps_in, :-1]
         for i in range(len(sequences)-n_steps_in-n_steps_out+1)]
    y = [sequences[i+n_steps_in:i+n_steps_in+n_steps_out, -1]
         for i in range(len(sequences)-n_steps_in-n_steps_out+1)]
    return np.array(X, dtype=float), np.array(y, dtype=float)


if __name__ == "__main__":

    main()



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import StandardScaler
from sklearn.tree import DecisionTreeRegressor

# from func.helpers import ptabulate


# def get_columns_of_dtype(df, dty=None, data=False):
#     """get columns of dtype included in dty list"""
#     dty_dict = {
#         'obj': ['object'],
#         'num': ['float', 'int', 'int64', 'float64', 'int32', 'float32', 'int16', 'float16', 'int8'],
#         'bool': ['bool'],
#         'cat': ['category'],
#         'date': ['datetime64']
#     }

#     if not dty:  # if no type is specified, default to object type
#         dty = ["obj"]
#     # convert single string to list if necessary
#     if isinstance(dty, str):
#         dty = [dty]

#     # create a list of all the types specified in the argument
#     dty_list = []
#     _ = [dty_list.extend(dty_dict[t]) for t in dty if t in dty_dict]

#     return df.copy().select_dtypes(include=dty_list) if data else list(df.select_dtypes(include=dty_list).columns)


# def correlate_a_single_df_column(df, column):
#     """Correlate a single column with all other columns
#     Args:
#         df (pd.DataFrame): dataframe to correlate
#         column (str): column to correlate
#     Returns:
#         pd.DataFrame: correlation table
#     """

#     df = df.copy()
#     if len(df) > 20000:
#         df = df.sample(n=20000, random_state=42)
#     if isinstance(column, list):
#         column = column[0]
#     num_cols = get_columns_of_dtype(df, dty='num')
#     dfc = pd.DataFrame(df[num_cols].corrwith(
#         df[column], method='pearson')).reset_index(drop=False)
#     dfc.columns = ['feature', 'corr']
#     dfc['corr_abs'] = dfc['corr'].abs()
#     dfc = dfc.sort_values(
#         'corr_abs', ascending=False).iloc[1:].dropna().reset_index(drop=True)
#     return dfc


# def plot_all_histograms_from_a_df(df, title=None):
#     """plots all histograms from a dataframe df in a single figure
#     Seaborn is too slow for this task
#     """
#     if title is None:
#         title = 'Histograms of Features'
#     cols = 5
#     rows = len(df.columns)//cols + 1
#     fig = plt.figure(figsize=(cols*4, rows*2))

#     # for each column, plot a histogram
#     for i, col in enumerate(df.columns):
#         _ = fig.add_subplot(rows, cols, i+1)
#         plt.hist(x=df[col], bins=30)
#         plt.title(col)

#     fig.suptitle(title, fontsize=20, y=1.005)
#     fig.tight_layout()
#     plt.show()


# def plot_all_boxplot_from_a_df(df, title=None):
#     """ Plot Boxplots of Features from dataframe df"""
#     if title is None:
#         title = 'Boxplots of Features'
#     _, axs = plt.subplots(1, 1, figsize=(20, 4))
#     # Boxplot
#     sns.boxplot(data=df, ax=axs)
#     plt.suptitle(title, fontsize=20, y=1.005)
#     plt.tight_layout()
#     plt.show()


# def do_correlation_matrix(df, plot=False):
#     """Correlation matrix of a dataframe"""

#     # Get columns of numeric types
#     numeric_cols = get_columns_of_dtype(df, dty=['num'])
#     # Calculate the correlation matrix only for numeric columns
#     corr = df[numeric_cols].corr(method='pearson')
#     if plot:
#         # Plot the correlation matrix using a heatmap
#         sns.set_style("darkgrid")
#         plt.figure(figsize=(15, 10))
#         sns.heatmap(corr, annot=True)
#         plt.show()
#     return corr


# def fix_na(df):
#     for c in df.columns:
#         if df[c].dtype != 'object':
#             df[c] = df[c].replace([np.inf, -np.inf], np.NaN)
#             df[c] = df[c].interpolate(method='linear', limit_direction='both')
#     return df


# def calc_feat_importance(df, target):
#     """Calculate feature importance using logistic regression
#     from a dataframe and a target variable
#     Returns a dataframe with the features and importance scores
#     """

#     df = fix_na(df)

#     # Get columns of numeric types
#     cols_to_use = get_columns_of_dtype(df, dty=['num'])

#     # Creates X and y
#     X = df[cols_to_use].copy().drop(columns=target).copy()
#     y = df[target].copy()

#     dt = DecisionTreeRegressor(max_features=1.0, random_state=123)
#     # Fit the model with X and y
#     dt.fit(X, y)
#     importances = dt.feature_importances_

#     # Create a dataframe with the features and importance scores
#     feat_importance = pd.DataFrame({
#         'feature': X.columns,
#         'importance': importances
#     })
#     feat_importance['importance'] = feat_importance['importance'].apply(
#         lambda x: round(x, 4))

#     # Sort the features by importance score
#     feat_importance = feat_importance.sort_values(
#         'importance', ascending=False).reset_index(drop=True)
#     return feat_importance


# def df_explore_missing_values(df):
#     """Explore missing values in a dataframe"""
#     features_with_nan = []
#     nan_count = df.isna().sum()
#     nan_count = nan_count[nan_count > 0]
#     if len(nan_count) > 0:
#         # Sorting by feature name
#         nan_count = nan_count.reset_index(drop=False).rename(
#             columns={'index': 'feature', 0: 'nans'}).sort_values(by='feature', ascending=True)
#         # % of missing values
#         nan_count['nan_p'] = nan_count['nans'] / len(df)
#         # Unique values
#         nan_count['unique_values'] = [len(df[c].unique())
#                                       for c in nan_count.feature]
#         # Data type
#         nan_count['dtype'] = df[nan_count.feature].dtype.str
#         # Save features with nan values for further analysis
#         features_with_nan = sorted(nan_count.feature)

#     return features_with_nan


# def vbar_plot(x, y, title, xlabel=None, ylabel=None, rotate=False, palette='Blues_r', figsize=(20, 5)):
#     sns.set_theme(style="whitegrid")
#     plt.figure(figsize=figsize)
#     sns.barplot(x=x, y=y, palette=palette)
#     if xlabel:
#         plt.xlabel(xlabel)
#     if ylabel:
#         plt.ylabel(ylabel)
#     plt.title(title)
#     # Add data labels to the bars
#     for i, value in enumerate(y):
#         plt.text(i, value, f'{value:.4f}', ha='center', va='bottom')
#     if rotate:
#         plt.xticks(rotation=45)

#     ylim_min = min(y) - 0.1 * max(y)
#     ylim_max = max(y) + 0.1 * max(y)
#     plt.ylim(ylim_min, ylim_max)

#     plt.tight_layout()
#     plt.show()


# def vbar_2_plots(df, title, xlabel=None, ylabel=None, rotate=False):
#     cols = df.columns.tolist()

#     # Set the width of each bar
#     bar_width = 0.35

#     # Calculate the positions of the bars on the x-axis
#     r1 = np.arange(len(df))
#     r2 = [x + bar_width for x in r1]

#     plt.figure(figsize=(10, 6))

#     # Plot the bars for mape_0
#     plt.bar(r1, df[cols[1]], width=bar_width, label=cols[1],
#             color='#536DFE', edgecolor='black')

#     # Plot the bars for mape_hyper
#     plt.bar(r2, df[cols[2]], width=bar_width,
#             label=cols[2], color='#3F51B5', edgecolor='black')
#     if xlabel is not None:
#         plt.xlabel(xlabel)
#     if ylabel is not None:
#         plt.ylabel(ylabel)

#     plt.title(title)

#     # Add data labels to the bars
#     for i, row in df.iterrows():
#         plt.text(i, row[cols[1]],
#                  f'{row[cols[1]]:.4f}', ha='center', va='bottom')
#         plt.text(i + bar_width, row[cols[2]],
#                  f'{row[cols[2]]:.4f}', ha='center', va='bottom')

#     # Set the x-axis tick positions and labels
#     if rotate:
#         plt.xticks([r + bar_width/2 for r in range(len(df))],
#                    df[cols[0]], rotation=45)
#     else:
#         plt.xticks([r + bar_width/2 for r in range(len(df))],
#                    df[cols[0]])

#     plt.legend()
#     plt.tight_layout()
#     plt.show()



#!/usr/bin/env python
# coding: utf-8
################################################################################
"""
NEURAL NETWORK CLASS MODULE

pretrained_glove: load embedding from pretrained glove
custom_standardization: layers preprocessing for text data
text_prep_layer: text preprocessing layer
normalize_layer: normalize data layer, 0-255
image_prep_layer: image preprocessing layer
rnn_layer: RNN layer
deep_layer: DNN layer


The ideal machine learning model is end-to-end
The ideal model should expect as input something as close as possible to raw data:
and a text model should accept strings of utf-8 characters

"""
################################################################################

import tensorflow as tf
import tensorflow_text as text  # need to keep this line to work
from text import bert
#import neural_structured_learning as nsl # TODO ?

class NETWORK(tf.keras.Model):
    def __init__(self, X, Y, d_params, h_params):
        super(NETWORK, self).__init__()

        #if "GPU" not in str(tf.config.list_physical_devices()):  # TODO: check for gpu on GPU compati.
        #    raise Exception("Error: GPU not being used")

        self.columns = X.columns
        self.F, self.O = 0, Y.shape[-1]

        """ INPUT LAYER """
        self.dtypes_x = {k: val["type"] for k, val in d_params["x"].items()}
        self.dtypes_y = {k: val["type"] for k, val in d_params["y"].items()}
        self.dtypes = list(zip(X.dtypes.index, X.dtypes))  # get dtypes

        # Use Dtypes to Create Mixed Data
        input_els, encoded_els = [], []
        for k, dtype in self.dtypes_x.items():  # structure the data
            self.F += 1
            if dtype == "CATEGORY":  # categorical data (str, int, bool)
                input_els.append(tf.keras.Input(shape=(1,), dtype=tf.int32))
                n = len(X[k].unique())
                x = tf.keras.layers.Embedding(n+1, 1)(input_els[-1])  # takes only 999 size
                e = tf.keras.layers.Flatten()(x)
                self.F += n
            elif dtype == "NUMBER":  # continuous data (float, int)
                input_els.append(tf.keras.Input(shape=(1,), dtype=tf.float32))
                e = tf.keras.layers.BatchNormalization()(input_els[-1])  # TODO verify int cont. works like this
            elif dtype == "STRING":  # textual data (str)
                input_els.append(tf.keras.Input(shape=(1,), dtype=tf.string))
                encoder = bert.NETWORK(h_params, Y.shape[-1])
                e = encoder(input_els[-1])
            #elif dtype == "STRUCTURED":  # TODO: pass structured as var or
            #    #graph_config = nsl.configs.GraphRegConfig(neighbor_config=nsl.configs.GraphNeighborConfig(max_neighbors=1)) 
            #    #graph_model = nsl.keras.GraphRegularization(base_model, graph_config)
            #    
            else:
                raise Exception(dtype)
            encoded_els.append(e)

        x = tf.keras.layers.concatenate(encoded_els)  # layers are now merged

        # Deep Layer
        n = x.shape.as_list()[-1]
        for i in range(h_params["deep_lyrs"]):
            x = tf.keras.layers.Dense(n, activation="relu")(x)

        # Add Regularization For Small Datasets
        # x = tf.keras.layers.Dropout(0.5)(x)

        """ TRANSLATION LAYER """

        # Output Layer TODO different activations
        out = tf.keras.layers.Dense(self.O, activation="sigmoid")(x)

        self.model = tf.keras.Model(inputs=input_els, outputs=[out])

        """ LOSS """

        # Check if Binary  TODO multiple types of output & cost
        binary = len(list(set(Y.values.ravel('K')))) == 2

        if binary and (self.O == 1):  # binary
            self.loss, self.metric = "binary_crossentropy", [tf.keras.metrics.BinaryAccuracy()]
        elif binary and (self.O > 1):  # binary
            self.loss, self.metric = "categorical_crossentropy", [tf.keras.metrics.CategoricalAccuracy()]
        elif not binary:
            self.loss, self.metric = "mse", [tf.keras.metrics.MeanSquaredError()]
        else:
            raise Exception("Error: loss not setup in network.py")

    # @tf.function  # (input_signature=(tf.TensorSpec(shape=(None, 12), dtype=tf.float32),))
    # def f(self, X):
    #     return self.model(tf.unstack(tf.transpose(X)))

    def call(self, x, training=False):
        if training:
            return self.model(x)
        else:
            try:
                print("eagerly == false", x.numpy())
            except:
                pass
            return self.model(x)


    #def get_config(self):
    #    return {"hidden_units": self.hidden_units}
    #
    #@classmethod
    #def from_config(cls, config):
    #    return cls(**config)



# Dependent Variable (Target)

BASE_DB_DICT_CONF = {
    'remove_hours': True,
    'forward_hours': 0,
    "delta": list(range(1, 8)),
    "lag": list(range(1, 8)),
    "add_nld": True,
    "dropna": True,
    "forward": 1,
    "do_scale": True,
    "do_pca": True,
    "scale_y": True,
    "reduce": True,
    "scaler_x": "std",
    "scaler_y": "std",
    "resample": "H",
    "add_meanstd": True,
}


# def ptabulate(df):
#     # print(tabulate(df, headers='keys', tablefmt='github'))
#     print()


# def print_as_title(text):
#     print()
#     print(text)
#     print('='*len(text))



import numpy as np
import pandas as pd

def cap_outliers(s):
    if isinstance(s, pd.Series):
        s_ = s.copy()
        s_ = s_.dropna().tolist()
        s2 = pd.Series(index=s.index)
    else:
        s_ = s
        s2 = []

    Q1 = np.percentile(s_, 25)
    Q3 = np.percentile(s_, 75)
    IQR = Q3 - Q1
    upper = Q3 + 1.5 * IQR
    lower = Q1 - 1.5 * IQR

    for i, x in enumerate(s):
        if pd.notnull(x):
            if x > upper:
                s2[i] = upper
            elif x < lower:
                s2[i] = lower
            else:
                s2[i] = x
        else:
            s2[i] = np.nan

    return s2


def remove_redundant_features(df, target, threshold=0.9, drop=True):
    """Removes redundant features based on the Pearson correlation coefficient.
    Args:
        df (pandas.DataFrame): The dataset containing the features.
        target (str): The name of the target column.
        threshold (float, optional): The threshold above which the correlation coefficient is considered redundant. Defaults to 0.9.
        drop (bool, optional): Whether to drop the redundant features or not. Defaults to True.
    Returns:
        df (pandas.DataFrame): The dataset with the redundant features dropped.
        cols_to_drop (list): The columns that were dropped.
    """

    # Check if the target column is in the dataframe
    if target not in df.columns:
        raise KeyError(
            f'The target column "{target}" does not exist in the dataframe.')

    # Copy the target column
    dependent_series = df[target].copy()

    # Sample the dataframe to reduce the time to calculate the correlation matrix
    dfs = df.copy().sample(min([10_000, len(df)]), random_state=42)

    # Calculate the Pearson correlation matrix and Convert the matrix to a dictionary
    corr_matrix = dfs.corr(method='pearson').abs()
    corr_dict = corr_matrix[[target]].reset_index(
        drop=False).to_dict('records')
    corr_dict = {k['index']: k[target] for k in corr_dict}

    # Create a list to store the columns to drop
    cols_to_drop = []
    dict_cols_to_drop = []
    for i, c in enumerate(corr_matrix.columns):
        for y, f in enumerate(corr_matrix[c].index):
            # Skip the diagonal and lower triangle elements
            if y >= i:
                continue
            # Check if the correlation coefficient is above the threshold and append the column with the lowest corr coef to the list
            if corr_matrix[c].loc[f] > threshold:
                dict_cols_to_drop.append(
                    {"f0": c, "f1": f, "corr_f0f1": corr_matrix[c].loc[f], "f0_corr": corr_dict[c], "f1_corr": corr_dict[f], "remove": c if corr_dict[c] < corr_dict[f] else f})
                cols_to_drop.append(c if corr_dict[c] < corr_dict[f] else f)

    # Remove duplicate columns and the target column
    cols_to_drop = [d for d in list(set(cols_to_drop)) if d != target]

    if drop:
        df = df.drop(cols_to_drop, axis=1)
        df[target] = dependent_series

    return df, cols_to_drop, dict_cols_to_drop


def generalized_fisher_score(dataset, target_column):
    """Calculate the Generalized Fisher Score for each feature in the dataset."""
    feature_columns = [col for col in dataset.columns if col != target_column]

    # Calculate mean and standard deviation for each feature and target
    means = dataset.mean()
    stds = dataset.std()

    gfs_scores = {}

    for feature in feature_columns:
        # Calculate the covariance matrix
        cov_matrix = np.cov(dataset[[feature, target_column]].T)

        # Calculate the Fisher score
        fisher_score = (cov_matrix[0, 1] ** 2) / \
            (stds[feature] * stds[target_column])

        # Calculate the Generalized Fisher Score
        gfs_score = (means[feature] * means[target_column] *
                     fisher_score) / (stds[feature] * stds[target_column])

        # Store the GFS score
        gfs_scores[feature] = gfs_score

    # GFS dict to DataFrame sorted by GFS score in descending order
    gfs_scores = pd.DataFrame.from_dict(gfs_scores, orient='index', columns=[
                                        'GFS'])
    gfs_scores['GFS'] = gfs_scores['GFS'].apply(lambda x: round(x, 4))
    gfs_scores = gfs_scores.sort_values(
        by='GFS', ascending=False)
    return gfs_scores



import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from matplotlib.colors import ListedColormap
from PIL import Image
import random
import seaborn as sns
import numpy as np
from scipy.interpolate import interp1d
import matplotlib.ticker as ticker
# import squarify

colors =["#1D2026","#393939","#666666","#BDBDBD","#E0E0E0","#F2F2F2","#FD93B1","#FC2964","#FEC9D8","#FD5F8B","#0095FF","#E6F5FF","#5C33F6","#8566F8",]

def set_graph_dark(fig,ax,**optional_args):
    theme ='black'
    if 'theme' in optional_args:
        theme= optional_args['theme']
    if theme == 'black':
        ax.set_facecolor('black')
        fig.patch.set_facecolor('black')
        ax.spines['top'].set_color('white')
        ax.spines['right'].set_color('white')
        ax.spines['bottom'].set_color('white')
        ax.spines['left'].set_color('white')
        ax.yaxis.set_tick_params(colors='white')
        ax.xaxis.set_tick_params(colors='white') 
        ax.xaxis.label.set_color('white') 
        ax.yaxis.label.set_color('white')
        ax.set_title('',color='white')

    if('box_plots' in optional_args):
        for element in ['caps', 'whiskers']:
            for box in optional_args['box_plots'][element]:
                box.set(color='white')


def line_chart(xAxisvalue, predicatedLine, **optional_params):
    x = np.arange(len(xAxisvalue))
    x_smooth = np.linspace(0, len(xAxisvalue) - 1, 100)
    f_actual = interp1d(x, optional_params['actualLine'], kind='cubic')
    f_predicted = interp1d(x, predicatedLine, kind='cubic')
    actualLine_smooth = f_actual(x_smooth)
    predicatedLine_smooth = f_predicted(x_smooth)
    fig, ax = plt.subplots(figsize=(16, 9))
    ax.grid(True, linestyle='--', alpha=0.5)
    ax.plot(x_smooth, actualLine_smooth, color="#FF8080", label='Actual')
    ax.plot(x_smooth, predicatedLine_smooth, color="#80A0FF", label='Predicted')

    theme = 'black'
    if 'theme' in optional_params:
        theme = optional_params['theme']
        if theme == 'black':
            plt.rcParams['text.color'] = 'white'
            plt.rcParams['axes.labelcolor'] = 'white'
            plt.rcParams['axes.edgecolor'] = 'white'
            plt.rcParams['xtick.color'] = 'white'
            plt.rcParams['ytick.color'] = 'white'
            ax.spines['bottom'].set_color('white')
            ax.spines['left'].set_color('white')
            ax.spines['top'].set_color('white')
            ax.spines['right'].set_color('white')

    legend = ax.legend(bbox_to_anchor=(1.02, 1), facecolor='black' if theme == 'black' else 'white', edgecolor='black' if theme == 'white' else 'white')
    legend.get_texts()[0].set_color('black' if theme == 'white' else 'white')
    legend.get_texts()[1].set_color('black' if theme == 'white' else 'white')
    legend.get_frame().set_facecolor('black' if theme == 'black' else 'white')
    legend.get_frame().set_edgecolor('white' if theme == 'black' else 'black')
    ax.xaxis.set_major_locator(plt.IndexLocator(base=1, offset=0))
    ax.set_xticklabels(xAxisvalue, color='black' if theme == 'white' else 'white')
    ax.yaxis.set_tick_params(colors='black' if theme == 'white' else 'white')
    ax.set_facecolor('black' if theme == 'black' else 'white')
    fig.patch.set_facecolor('black' if theme == 'black' else 'white')
    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'], color='white' if theme == 'black' else 'black')
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'], color='white' if theme == 'black' else 'black')
    if ('title' in optional_params):
        ax.set_title(optional_params['title'], color='white' if theme == 'black' else 'black')
    name = "line_chart"
    plt.savefig(name + '.png', facecolor='black' if theme == 'black' else 'white', transparent=True)
    im = Image.open(name + '.png')
    print("Image saved:", name + '.png')
    im.show()


# xAxisvalue = ['8:00 PM', '9:00 PM', '10:00 PM', '12:00 PM', '1:00 AM', '3:00 AM', '3:00 AM', '4:00 AM', '5:00 AM',
#               '3:00 AM', '3:00 AM', '4:00 AM', '5:00 AM']
# xlabel='X-Label'
# predicatedLine = [30, 30, 33, 28, 34, 29, 33, 28, 34, 29, 28, 33, 32]
# actualLine = [30, 30, 29, 33, 32, 29, 29, 33, 32, 29, 30, 31, 37]
# theme = "black"
# line_chart(xAxisvalue, predicatedLine, xlabel=xlabel, actualLine=actualLine, theme=theme)




def horizontal_waterfall(categories, values, **optional_params):
    fig, ax = plt.subplots()
    theme = 'black'
    if 'theme' in optional_params:
        theme = optional_params['theme']
    categories = [c for c, v in zip(categories, values) if v is not None]
    values = [v for v in values if v is not None]
    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'], fontsize=12, color='white' if theme == 'black' else 'black')
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'], fontsize=8, color='white' if theme == 'black' else 'black')
    if 'title' in optional_params:
        title = optional_params['title']
        theme = optional_params.get('theme', 'black')
        title_color = 'white' if theme == 'black' else 'black'
        ax.set_title(title, fontsize=15, color=title_color)

    bars = ax.barh(categories, values, align='center')

    if theme == 'black':
        ax.spines['top'].set_visible(True)
        ax.spines['right'].set_visible(True)
        ax.spines['bottom'].set_visible(True)
        ax.spines['left'].set_visible(True)
        ax.spines['bottom'].set_color('white')
        ax.spines['left'].set_color('white')
        ax.spines['top'].set_color('white')
        ax.spines['right'].set_color('white')
        ax.spines['bottom'].set_linewidth(0.8)
        ax.spines['left'].set_linewidth(0.8)
        ax.spines['top'].set_linewidth(0.8)
        ax.spines['right'].set_linewidth(0.8)

        text_color = 'white'
        bar_color = '#0095FF'
        fig.set_facecolor('black')
        ax.grid(axis='x', color='white', linestyle='--', alpha=0.5)
        ax.set_axisbelow(True)
    else:
        text_color = 'black'
        bar_color = '#0095FF'
        fig.set_facecolor('white')
        ax.grid(axis='x', color='black', linestyle='--', alpha=0.5)
        ax.set_axisbelow(True)

    for bar, value in zip(bars, values):
        if value < 0:
            bar.set_color('#FC2964')
        else:
            bar.set_color(bar_color)

    ax.set_yticklabels(categories, fontsize=8, color=text_color)
    ax.set_xticklabels(ax.get_xticks(), fontsize=8, color=text_color)

    ax.set_facecolor('white' if theme == 'white' else 'black')

    name = str(random.random())
    plt.savefig(name + '.png', facecolor='white' if theme == 'white' else 'black')
    im = Image.open(name + '.png')
    print("Image saved:", name + '.png')
    im.show()

    if 'theme' in optional_params:
        print("Theme:", optional_params['theme'])

# categories = ['Burger', 'Pizza', 'Sandwich', 'Pasta', 'Momos', 'Garlic']
# values = [-10, -5, -4, 4, 5, 10]
# title = 'Horizontal Waterfall'
# theme = 'white'
# xlabel='xlabel'
# horizontal_waterfall(categories, values, xlabel=xlabel, title=title, theme=theme)

def bar_chart( categories, series, **optional_params):
    fig, ax = plt.subplots()
    bar_labels =[]
    if('bar_labels' in optional_params):
        bar_labels = optional_params['bar_labels']
    else:
        for i in range(0, len(categories)+ 1,1):
            bar_labels.append(i)
    colors=["#1D2026","#FC2964","#0095FF","#5C33F6","#393939","#FD93B1","#E6F5FF","#8566F8","#666666","#FEC9D8","#BDBDBD","#FD5F8B","#E0E0E0","#F2F2F2"]
    bar_colors =[]
    if('bar_colors' in optional_params):
        bar_colors = optional_params['bar_colors']
    else:
        for i in range(0, len(categories)+ 1,1):
            if(i <= len(colors) -1):
                bar_colors.append(colors[i])
    data = series
    data = list(map(lambda a, i: {'index': i, 'value': a}, data, range(len(data))))
    indexes = sorted(data, key=lambda x: x['value'])
    indexes = [a['index'] for a in indexes]
    values = [a['value'] for a in sorted(data, key=lambda x: x['value'])]
    cat = []
    labels =[]
    colors = []
    for d in indexes:
        cat.append(categories[d])
        labels.append(bar_labels[d])
        colors.append(bar_colors[d])
    ax.barh(cat, values,label=labels ,color=colors)
    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'])
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])
    if ('legend' in optional_params):
        ax.legend(title=optional_params['legend'])
    name = random.random()
    # plt.savefig( str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png')
    print("Image saved: ", str(name) + '.png')
    im.show() 
  

def column_chart( categories, series, **optional_params):
    fig, ax = plt.subplots()
    bar_labels =[]
    if('bar_labels' in optional_params):
        bar_labels = optional_params['bar_labels']
    else:
        for i in range(0, len(categories)+ 1,1):
            bar_labels.append(i)

    colors=["#1D2026","#FC2964","#0095FF","#5C33F6","#393939","#FD93B1","#E6F5FF","#8566F8","#666666","#FEC9D8","#BDBDBD","#FD5F8B","#E0E0E0","#F2F2F2"]
    bar_colors =[]
    if('bar_colors' in optional_params):
        bar_colors = optional_params['bar_colors']
    else:
        for i in range(0, len(categories)+ 1,1):
            if(i <= len(colors) -1):
                bar_colors.append(colors[i])
    ax.bar(categories, series,label=bar_labels ,color=bar_colors)
    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'])
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])
    if ('legend' in optional_params):
        ax.legend(title=optional_params['legend'])
    name = random.random()
    # plt.savefig( str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png') 
    print("Image saved: ", str(name) + '.png')
    im.show() 
  
  

def stack_bar( categories, below,above, **optional_params):
    species = categories
    weight_counts = {
        "Below": np.array(below),
        "Above": np.array(above),
    }
    width = 0.5
    fig, ax = plt.subplots()
    bottom = np.zeros(3)
    index= 0
    color=["#0095FF","#E6F5FF"]
    if ('color' in optional_params):
        color= (optional_params['color'])
    for boolean, weight_count in weight_counts.items():
        p = ax.bar(species, weight_count, width, color = color[index], label=boolean, bottom=bottom)
        index = index+1
        bottom += weight_count
    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'])
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])
    if ('legend' in optional_params):
        ax.legend(title=optional_params['legend'])
    name = random.random()
    # plt.savefig( str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png') 
    print("Image saved: ", str(name) + '.png')
    im.show() 
  

def line_graph(categories, data, label,**optional_params):
    fig, ax = plt.subplots()
    colors=["#1D2026","#FC2964","#0095FF","#5C33F6","#393939","#FD93B1","#E6F5FF","#8566F8","#666666","#FEC9D8","#BDBDBD","#FD5F8B","#E0E0E0","#F2F2F2"]
    color = colors
    if ('color' in optional_params):
        color =(optional_params['color'])
    for i in range(0, len(data)):
        print(label[i])
        ax.plot(categories,data[i],color=  color[i] if i<len(color) else  colors[i], label=label[i])
    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'])
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])
    ax.legend()
    name = random.random()
    # plt.savefig( str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png') 
    print("Image saved: ", str(name) + '.png')
    im.show() 
  
def heat_graph( data,  **optional_params):
    fig, ax = plt.subplots()
    data= np.array(data)
    color = colors[0]
    if ('color' in optional_params):
        color =(optional_params['color'])
    cmap = mcolors.LinearSegmentedColormap.from_list('my_colormap', [color, color])
    sns.heatmap(data, annot=True, cmap=cmap, alpha=data)
    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'])
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])
    if ('legend' in optional_params):
        ax.legend(title=optional_params['legend'])
    name = random.random()
    # plt.savefig( str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png') 
    print("Image saved: ", str(name) + '.png')
    im.show()                  
    
def histogram_graph(data , bins ,**optional_params):
    fig, ax = plt.subplots()
    color = colors[10]
    if ('color' in optional_params):
        color =(optional_params['color'])
    plt.hist(data, bins, alpha=0.7, edgecolor='black',color=color)
    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'])
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])
    if ('legend' in optional_params):
        ax.legend(title=optional_params['legend'])
    values, counts = np.unique(data, return_counts=True)
    plot_color = colors[9]
    if ('plot_color' in optional_params):
        plot_color =(optional_params['plot_color'])
    plt.plot(values, counts, 'bo', markersize=8, color= plot_color)
    name = random.random()
    # plt.savefig( str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png') 
    print("Image saved: ", str(name) + '.png')
    im.show()        
  

  

  
def dot_graph(categories, data,**optional_params):
    fig, ax = plt.subplots() 
    color = colors[0]
    data = data
    data = list(map(lambda a, i: {'index': i, 'value': a}, data, range(len(data))))
    indexes = sorted(data, key=lambda x: x['value'])
    indexes = [a['index'] for a in indexes]
    values = [a['value'] for a in sorted(data, key=lambda x: x['value'])]
    cat = []
    for d in indexes:
        cat.append(categories[d])
    if ('color' in optional_params):
        color =(optional_params['color'])
    plt.scatter(cat,values,color=color)
    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'])
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])
    if ('legend' in optional_params):
        ax.legend(title=optional_params['legend'])
    name = random.random()
    # plt.savefig( str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png') 
    print("Image saved: ", str(name) + '.png')
    im.show()  


import random
import matplotlib.pyplot as plt
from PIL import Image

def vertical_waterfall(categories, values, **optional_params):
    fig, ax = plt.subplots()
    categories = [c for c, v in zip(categories, values) if v is not None]
    values = [v for v in values if v is not None]

    if 'title' in optional_params:
        title = optional_params['title']
        theme = optional_params.get('theme', 'black')
        title_color = 'white' if theme == 'black' else 'black'
        ax.set_title(title, fontsize=15, color=title_color)

    cumulative_sum = [sum(values[:i + 1]) for i in range(len(values))]
    bars = ax.bar(categories, values, align='center')

    ax.set_xlabel('Categories', fontsize=12, color='white')

    ax.set_ylabel('Value', fontsize=12, color='white')
   

    theme = 'black'
    if 'theme' in optional_params:
        theme = optional_params['theme']

    if theme == 'black':
        ax.spines['top'].set_visible(True)
        ax.spines['right'].set_visible(True)
        ax.spines['bottom'].set_visible(True)
        ax.spines['left'].set_visible(True)
        ax.spines['bottom'].set_color('white')
        ax.spines['left'].set_color('white')
        ax.spines['top'].set_color('white')
        ax.spines['right'].set_color('white')
        ax.spines['bottom'].set_linewidth(0.8)
        ax.spines['left'].set_linewidth(0.8)
        ax.spines['top'].set_linewidth(0.8)
        ax.spines['right'].set_linewidth(0.8)

        text_color = 'white'
        bar_color = '#0095FF'
        fig.set_facecolor('black')
        ax.grid(axis='y', color='white', linestyle='--', alpha=0.5)
        ax.set_axisbelow(True)
        ax.tick_params(axis='y', colors='white')  # Set y-axis tick color to white
    else:
        ax.spines['top'].set_visible(True)
        ax.spines['right'].set_visible(True)
        ax.spines['bottom'].set_visible(True)
        ax.spines['left'].set_visible(True)
        ax.spines['bottom'].set_color('black')
        ax.spines['left'].set_color('black')
        ax.spines['top'].set_color('black')
        ax.spines['right'].set_color('black')
        ax.spines['bottom'].set_linewidth(0.8)
        ax.spines['left'].set_linewidth(0.8)
        ax.spines['top'].set_linewidth(0.8)
        ax.spines['right'].set_linewidth(0.8)

        text_color = 'black'
        bar_color = '#0095FF'
        fig.set_facecolor('white')
        ax.grid(axis='y', color='black', linestyle='--', alpha=0.5)
        ax.set_axisbelow(True)
        ax.tick_params(axis='y', colors='black')  # Set y-axis tick color to black

    ax.set_xticks(range(len(categories)))  # Adjust the x-ticks positions
    ax.set_xticklabels(categories, rotation=45, fontsize=8, color=text_color)

    ax.set_facecolor('white' if theme == 'white' else 'black')
    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'])
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])
    if ('legend' in optional_params):
        ax.legend(title=optional_params['legend'])
    for bar, value in zip(bars, values):
        if value < 0:
            bar.set_color('#FC2964')
        else:
            bar.set_color(bar_color)

    name = str(random.random())
    plt.savefig(name + '.png', facecolor='white' if theme == 'white' else 'black')
    im = Image.open(name + '.png')
    print("Image saved:", name + '.png')
    im.show()

    # Show theme information
    if 'theme' in optional_params:
        print("Theme:", optional_params['theme'])


# categories = ['W1', 'W3', 'W5', 'W7', 'W9', 'W10']
# values = [10, 5, 4, -4, -5, -10]
# title = 'Vertical Waterfall'
# theme = 'black'
# vertical_waterfall(categories, values, title=title, theme=theme)

def baseline_graph(baseline, x_labels, series_colors,*series_data, **optional_params):
    fig, ax = plt.subplots()

    for i, series in enumerate(series_data):
        percentage_change_series = [(value - baseline[j]) / baseline[j] * 100 for j, value in enumerate(series)]
        color = series_colors[i % len(series_colors)]
        ax.plot(x_labels, percentage_change_series, label=f'Product {i+1}', marker='o', color=color)
    theme ='black'
    if 'theme' in optional_params:
        theme= optional_params['theme']
    ax.axhline(y=0,  color='black' if theme != 'black'else 'white', linestyle='dotted', label='Baseline 1')

    ax.grid(True, color='black' if theme != 'black'else 'white', linestyle='--', alpha=0.5)
    ax.set_title('Chg. from Baseline', fontdict={'fontsize': 15, 'color': 'black' if theme == 'black'else 'white'})
    legend = ax.legend(bbox_to_anchor=(1, 0.91), facecolor='black' if theme == 'black'else 'white')
    for text in legend.get_texts():
        text.set_color('black' if theme != 'black'else 'white')

    ax.set_xlabel('X Labels', fontdict={'fontsize': 12, 'color': 'black' if theme == 'black'else 'white'})
    ax.set_ylabel('Percentage Change', fontdict={'fontsize': 12, 'color':'black' if theme == 'black'else 'white'})
    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'], color='white' if theme == 'black' else 'black')
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'], color='white' if theme == 'black' else 'black')
    if ('title' in optional_params):
        ax.set_title(optional_params['title'], color='white' if theme == 'black' else 'black')
    if ('legend' in optional_params):
        ax.legend(title=optional_params['legend'])
    if theme == 'black':
        ax.spines['top'].set_visible(True)
        ax.spines['right'].set_visible(True)
        ax.spines['bottom'].set_color('white')
        ax.spines['left'].set_color('white')
        ax.spines['top'].set_color('white')
        ax.spines['right'].set_color('white')
        ax.spines['bottom'].set_linewidth(0.8)
        ax.spines['left'].set_linewidth(0.8)
        ax.tick_params(axis='x', colors='white')
        ax.tick_params(axis='y', colors='white')
        ax.set_facecolor('black')
        fig.patch.set_facecolor('black')

    name = str(random.random())
    plt.savefig(name + '.png', bbox_inches='tight', facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(name + '.png')
    print("Image saved:", name + '.png')
    im.show()


# baseline = [10, 12, 14, 16, 18]
# series1 = [11, 15, 12, 13, 20]
# series2 = [9, 11, 13, 15, 17]
# series3 = [13, 14, 10, 12, 11]
# x_labels = ['W1', 'W3', 'W5', 'W7', 'W9']
# colors = ["#FC2964", "#0095FF", "#5C33F6"]
# title='Title'
# theme = 'white'
# baseline_graph(baseline, x_labels, colors, series1, series2, series3, title=title, theme=theme)




def scatter_plot_graph(marker_shapes, x_values, series_data, colors, **optional_params):
    fig, ax = plt.subplots(figsize=(11, 9))
    np.random.seed(19680801)
    sizes = np.random.rand(len(x_values[0])) * 50 + 50

    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig,ax,theme=theme)
    for i, marker in enumerate(marker_shapes):
        x = x_values[i]
        y = series_data[i]
        color = colors[i % len(colors)]
        ax.scatter(x, y, s=sizes, alpha=0.5, marker=marker, color=color, label="Group {}".format(i+1))

    legend = ax.legend(bbox_to_anchor=(1, 0.9), facecolor='black' if theme == 'black' else 'white')

    if theme == 'black':
        for text in legend.get_texts():
            text.set_color('white')

    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'], fontdict={'fontsize': 12})   
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'],fontdict={'fontsize': 12}) 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'],fontdict={'fontsize': 15})

    name = str(random.random())
    plt.savefig(name + '.png',  bbox_inches='tight',facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(name + '.png')
    print("Image saved:", name + '.png')
    im.show()
    # plt.show()

# marker_shapes = ['s', 'o', 'D']
# x_values = [np.arange(0.15, 50.0, 2.0), np.arange(0.15, 50.0, 2.0), np.arange(0.15, 50.0, 2.0)]
# series_data = [
#     x_values[0] ** 1.3 + np.random.rand(*x_values[0].shape) * 30.0,
#     x_values[1] ** 1.5 + np.random.rand(*x_values[1].shape) * 40.0,
#     x_values[2] ** 1.2 + np.random.rand(*x_values[2].shape) * 20.0
# ]
# title='Title'
# colors = ["#FC2964", "#0095FF", "#5C33F6"]
# scatter_plot_graph(marker_shapes, x_values, series_data, colors,title=title, theme='white')

def stack_bar(categories, below, above, **optional_params):
    species = categories
    weight_counts = {
        "Below": np.array(below),
        "Above": np.array(above),
    }
    width = 0.5

    fig, ax = plt.subplots()

    bottom = np.zeros(len(categories))
    index = 0
    color = ["#FC2964", "#0095FF"]
    if 'color' in optional_params:
        color = optional_params['color']

    for boolean, weight_count in weight_counts.items():
        p = ax.bar(species, weight_count, width, color=color[index], label=boolean, bottom=bottom)
        index += 1
        bottom += weight_count

    if 'xlabel' in optional_params:
        ax.set_xlabel(optional_params['xlabel'], color='white')
    if 'ylabel' in optional_params:
        ax.set_ylabel(optional_params['ylabel'], color='white')
    if 'title' in optional_params:
        ax.set_title(optional_params['title'], color='white')
    if 'legend' in optional_params:
        legend = ax.legend(loc='lower center', title=optional_params['legend'], bbox_to_anchor=(0.50, -0.25), ncol=2)
        
    fig.tight_layout()

    theme ='black'
    if 'theme' in optional_params:
        theme= optional_params['theme']
    if theme == 'black':
        ax.spines['bottom'].set_color('white')
        ax.spines['left'].set_color('white')
        ax.spines['top'].set_color('white')
        ax.spines['right'].set_color('white')
        ax.spines['bottom'].set_linewidth(0.8)
        ax.spines['left'].set_linewidth(0.8)
        ax.spines['top'].set_linewidth(0.8)
        ax.spines['right'].set_linewidth(0.8)

        ax.tick_params(axis='x', colors='white')
        ax.tick_params(axis='y', colors='white')

        ax.set_facecolor('black')
        fig.patch.set_facecolor('black')

        legend.get_frame().set_facecolor('black')
        legend.get_title().set_color('white')
        for text in legend.get_texts():
            text.set_color('white')
    name = str(random.random())
    plt.savefig(name + '.png', facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(name + '.png')
    print("Image saved:", name + '.png')
    im.show()

# categories = ('2010','2012','2014','2016','2018','2020')
# below=[30,35,40,45,50,55]
# above=[70,70,45,50,55,60]
# title = 'Stack bar chart'
# ylabel='Title'
# legend='legends'
# stack_bar(categories,below,above,title=title,ylabel=ylabel,legend=legend,theme = 'black')

# series_data = [[ 50,  80, 20,  0,  20,  50], [ -10, -30, -10,  -80,  -30,  -10], [ -100, -60, -100,  -20,  0,  -100]]
# categories= ['Jan','Feb','Mar','Apr','May','Jun'] 
# title ='Popular items'
# ylabel ='In $'
# xlabel ='Month' # represent x axis title 
# label=[ 'Product 1', 'Product 2' , 'Product 3']
# line_graph(categories,series_data,label,title=title,ylabel = ylabel,xlabel = xlabel )

def heatmap_graph(y_axis_data, x_axis_data, colors,**optional_params):
    fig, ax = plt.subplots(figsize=(8, 8))
    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig,ax,theme=theme)

    data = np.random.rand(len(y_axis_data), len(x_axis_data))
    cmap = ListedColormap(colors)
    heatmap = ax.imshow(data, cmap=cmap, interpolation='nearest')
    ax.set_yticks(range(len(y_axis_data)))
    ax.set_xticks(range(len(x_axis_data)))
    cbar = plt.colorbar(heatmap)

    if theme == 'black':
        cbar.ax.tick_params(labelcolor='white')

    ax.set_xticklabels(ax.get_xticklabels(), ha='center')
    plt.subplots_adjust(bottom=0.2)

    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'], fontdict={'fontsize': 12})   
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'],fontdict={'fontsize': 12}) 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])

    name = str(random.random())
    plt.savefig(name + '.png', facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(name + '.png')
    print("Image saved:", name + '.png')
    im.show()
    # plt.show() 

# title='Heatmap'
# y_axis_data = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
# x_axis_data = [2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2017, 2019, 2021, 2023]
# colors = ["#1D2026", "#393939", "#666666", "#BDBDBD", "#E0E0E0", "#F2F2F2", "#FD93B1",
#           "#FC2964", "#FEC9D8", "#FD5F8B", "#0095FF", "#E6F5FF", "#5C33F6", "#8566F8"]
# heatmap_graph(y_axis_data, x_axis_data, colors,title=title,theme='black')

def heatmap2_graph(y_axis_data, x_axis_data, colors,dataFrom,dataTo,**optional_params):
    fig, ax = plt.subplots(figsize=(8, 8))
    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig,ax,theme=theme)

    data = np.random.randint(dataFrom, dataTo, size=(len(y_axis_data), len(x_axis_data)))
    cmap = ListedColormap(colors)
    heatmap = ax.imshow(data, cmap=cmap, interpolation='nearest')
    ax.set_yticks(range(len(y_axis_data)), y_axis_data)
    ax.set_xticks(range(len(x_axis_data)), x_axis_data)
    cbar = plt.colorbar(heatmap)

    if theme == 'black':
        cbar.ax.tick_params(labelcolor='white')
    for i in range(len(y_axis_data)):
        for j in range(len(x_axis_data)):
            ax.text(j, i, str(data[i, j]), ha='center', va='center', color='black')

    ax.set_xticklabels(ax.get_xticklabels(), ha='center')
    plt.subplots_adjust(bottom=0.2)

    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'], fontdict={'fontsize': 12})   
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'],fontdict={'fontsize': 12}) 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])

    name = random.random()
    plt.savefig(str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png') 
    print("Image saved: ", str(name) + '.png')
    im.show()
    # plt.show()

# title='Heatmap'
# dataFrom=1
# dataTo=100
# y_axis_data = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
# x_axis_data = [2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2017, 2019, 2021, 2023]
# colors = ["#1D2026", "#393939", "#666666", "#BDBDBD", "#E0E0E0", "#F2F2F2", "#FD93B1",
#           "#FC2964", "#FEC9D8", "#FD5F8B", "#0095FF", "#E6F5FF", "#5C33F6", "#8566F8"]
# heatmap2_graph(y_axis_data, x_axis_data, colors,dataFrom,dataTo,title=title,theme='black')

def boxPlot_graph(data, titles, colors, **optional_params):
    fig, ax = plt.subplots()
    ranges = [np.max(dataset) - np.min(dataset) for dataset in data]
    sorted_data = sorted(zip(data, ranges, titles, colors), key=lambda x: x[1], reverse=True)
    sorted_values, sorted_ranges, sorted_titles, sorted_colors = zip(*sorted_data)
    box_plot = ax.boxplot(sorted_values, patch_artist=True)

    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig,ax,box_plots = box_plot,theme=theme)
    for patch, color in zip(box_plot['boxes'], sorted_colors):
        patch.set(facecolor=color, linewidth=1)

    ax.set_xticklabels(sorted_titles)
    legend = ax.legend().remove()

    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'], fontdict={'fontsize': 12})   
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'],fontdict={'fontsize': 12}) 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])
    if('legend' in optional_params):
        legend.set_title(optional_params['legend'])

    name = random.random()
    plt.savefig(str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png') 
    print("Image saved: ", str(name) + '.png')
    im.show()
    # plt.show()

# data = [[25, 26, 27, 28, 29],
#         [20, 21, 22, 23, 24],
#         [10, 12, 14, 16, 18],
#         [10, 15, 20, 25, 30],
#         [20, 23, 26, 29, 32],
#         [22, 26, 29, 33, 36]]

# colors = ["#1D2026", "#393939", "#666666", "#BDBDBD", "#E0E0E0", "#F2F2F2", "#FD93B1",
#           "#FC2964", "#FEC9D8", "#FD5F8B", "#0095FF", "#E6F5FF", "#5C33F6", "#8566F8"]
# titles = ['Group 1', 'Group 2', 'Group 3', 'Group 4', 'Group 5', 'Group 6']
# xAxisTitle='Title'
# yAxisTitle='Title'
# boxPlot_graph(data, titles, colors,xlabel=xAxisTitle,ylabel=yAxisTitle,theme='white')

def boxPlot2_graph(data, titles, colors, **optional_params):
    fig, ax = plt.subplots()
    ranges = [np.max(dataset) - np.min(dataset) for dataset in data]
    sorted_data = sorted(zip(data, ranges, titles, colors), key=lambda x: x[1], reverse=True)
    sorted_values, sorted_ranges, sorted_titles, sorted_colors = zip(*sorted_data)
    box_plot = ax.boxplot(sorted_values, patch_artist=True)

    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig,ax,box_plots = box_plot,theme=theme)
    for patch, color in zip(box_plot['boxes'], sorted_colors):
        patch.set(facecolor=color, alpha=0.5, edgecolor=color, linewidth=1)

    ax.set_xticklabels(sorted_titles)
    legend = ax.legend().remove()

    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'], fontdict={'fontsize': 12})   
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'],fontdict={'fontsize': 12}) 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])
    if('legend' in optional_params):
        legend.set_title(optional_params['legend'])

    name = random.random()
    plt.savefig(str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png') 
    print("Image saved: ", str(name) + '.png')
    im.show()
    # plt.show()

# data = [[25, 26, 27, 28, 29],
#         [20, 21, 22, 23, 24],
#         [10, 12, 14, 16, 18],
#         [10, 15, 20, 25, 30],
#         [20, 23, 26, 29, 32],
#         [22, 26, 29, 33, 36]]

# colors = ["#FEC9D8","#FD5F8B","#0095FF","#1D2026","#5C33F6","#8566F8"]
# titles = ['Group 1', 'Group 2', 'Group 3', 'Group 4', 'Group 5', 'Group 6']
# xAxisTitle = 'Title'
# yAxisTitle = 'Title'
# boxPlot2_graph(data, titles, colors, xlabel=xAxisTitle, ylabel=yAxisTitle,theme='black')

def boxPlot3_graph(data, colors,**optional_params):
    fig, ax = plt.subplots(figsize=(10, 6))
    box_plots = ax.boxplot(data, patch_artist=True, vert=True, widths=0.5)

    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig, ax, box_plots=box_plots,theme=theme)
    for patch, color in zip(box_plots['boxes'], colors[:len(data)]):
        patch.set_facecolor(color)
        patch.set_edgecolor('white')

    legend_handles = [plt.Rectangle((0, 0), 1, 1, color=color) for color in colors[:len(data)]]
    legend_labels = [f'{i+1}' for i in range(len(data))]
    legend = ax.legend(legend_handles, legend_labels, loc='upper right', bbox_to_anchor=(1.1, 1), facecolor='black' if theme=='black' else 'white')

    if theme == 'black':
        for text in legend.get_texts():
            text.set_color('white')
    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])   
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel']) 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])
    if('legend' in optional_params):
        legend.set_title(optional_params['legend'])

    name = str(random.random())
    plt.savefig(name + '.png', facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(name + '.png')
    print("Image saved:", name + '.png')
    im.show()
    # plt.show()

# title = "Box Plot"
# xAxisTitle = "Legend"
# yAxisTitle = "Title"
# legend='Legend'
# colors = ["#1D2026", "#393939", "#666666", "#BDBDBD", "#E0E0E0", "#F2F2F2",
#           "#FD93B1", "#FC2964", "#FEC9D8"]
# data=[[10, 15, 20, 25, 30],
#       [20, 25, 29, 33, 35],
#       [22, 26, 29, 33, 36],
#       [22, 26, 29, 33, 36],
#       [22, 26, 29, 33, 36],
#       [20, 23, 26, 29, 32],
#       [20, 23, 26, 29, 32],
#       [10, 12, 14, 16, 18],
#       [20, 21, 22, 23, 24],
#       [20, 21, 22, 23, 24],
#       [25, 26, 27, 28, 29],
#       [20, 21, 22, 23, 24]
# ] 
# boxPlot3_graph(data, colors, legend=legend,title=title, xlabel=xAxisTitle, ylabel=yAxisTitle,theme='black')

def horizontal_waterfall(categories, values, **optional_params):
    fig, ax = plt.subplots()

    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig,ax,theme=theme)
        ax.grid(axis='x', color='white', linestyle='--', alpha=0.5)
    else:
        ax.grid(axis='x', color='black', linestyle='--', alpha=0.5)

    categories = [c for c, v in zip(categories, values) if v is not None]
    values = [v for v in values if v is not None]
    bars = ax.barh(categories, values, align='center')
    ax.set_axisbelow(True)
    ax.set_yticklabels(categories, fontdict={'fontsize': 8})
    ax.set_xticklabels(ax.get_xticks(), fontdict={'fontsize': 8})

    for bar, value in zip(bars, values):
        if value < 0:
            bar.set_color('#FC2964')
        else:
            bar.set_color('#0095FF')

    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'],fontdict={'fontsize': 12})   
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'],fontdict={'fontsize': 8}) 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'],fontdict={'fontsize': 15})

    name = str(random.random())
    plt.savefig(name + '.png', facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(name + '.png')
    print("Image saved:", name + '.png')
    im.show()
    # plt.show()

# categories = ['Burger', 'Pizza', 'sandwich', 'pasta', 'momos', 'garlic']
# values = [-10, -5, -4, 4, 5, 10]
# title = 'Secondary text'
# horizontal_waterfall(categories, values, title=title,theme='white')

def nextItemPrediction(categories, values, **optional_params):
    fig, ax = plt.subplots()

    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig,ax,theme=theme)
        ax.grid(axis='x', color='white', linestyle='--', alpha=0.5)
    else:
        ax.grid(axis='x', color='black', linestyle='--', alpha=0.5)

    categories = [c for c, v in zip(categories, values) if v is not None]
    values = [v for v in values if v is not None]
    bars = ax.barh(categories, values, align='center')
    ax.set_axisbelow(True)
    ax.set_yticklabels(categories, fontdict={'fontsize': 8})
    ax.set_xticklabels(ax.get_xticks(), fontdict={'fontsize': 8})

    for bar, value in zip(bars, values):
        if value < 0:
            bar.set_color('#FC2964')
        else:
            bar.set_color('#0095FF')

    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'],fontdict={'fontsize': 12})   
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'],fontdict={'fontsize': 8}) 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'],fontdict={'fontsize': 15})

    name = str(random.random())
    plt.savefig(name + '.png', facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(name + '.png')
    print("Image saved:", name + '.png')
    im.show()
    # plt.show()

# categories = ['Burger', 'Pizza', 'sandwich', 'pasta', 'momos', 'garlic']
# values = [-10, -5, -4, 4, 5, 10]
# title = 'Secondary text'
# nextItemPrediction(categories, values, title=title,theme='black')

def line_chart(xAxisvalue, predicatedLine, actualLine,**optional_params):
    fig, ax = plt.subplots(figsize=(16, 9))

    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig,ax,theme=theme)

    x = np.arange(len(xAxisvalue))
    x_smooth = np.linspace(0, len(xAxisvalue) - 1, 100)
    f_actual = interp1d(x, actualLine, kind='cubic')
    f_predicted = interp1d(x, predicatedLine, kind='cubic')
    actualLine_smooth = f_actual(x_smooth)
    predicatedLine_smooth = f_predicted(x_smooth)
    ax.grid(True, linestyle='--', alpha=0.5) 
    ax.plot(x_smooth, actualLine_smooth, color="#FF8080", label='Actual')  
    ax.plot(x_smooth, predicatedLine_smooth, color="#80A0FF", label='Predicted')
    legend = ax.legend(bbox_to_anchor=(1.02, 1), facecolor='black' if theme == 'black' else 'white')
    ax.xaxis.set_major_locator(plt.IndexLocator(base=1, offset=0))

    if theme=='black':
        legend.get_texts()[0].set_color('white') 
        legend.get_texts()[1].set_color('white')
        ax.set_xticklabels(xAxisvalue, color='white')
    else:
        ax.set_xticklabels(xAxisvalue, color='black')

    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'], color='white' if theme == 'black' else 'black')   
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'], color='white' if theme == 'black' else 'black') 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'],color='white' if theme == 'black' else 'black')

    name = str(random.random())
    plt.savefig(name + '.png', facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(name + '.png')
    print("Image saved:", name + '.png')
    im.show()
    # plt.show()

# xAxisvalue = ['8:00 PM', '9:00 PM', '10:00 PM', '12:00 PM', '1:00 AM', '3:00 AM', '3:00 AM', '4:00 AM', '5:00 AM','3:00 AM', '3:00 AM', '4:00 AM', '5:00 AM']
# title='Title'
# predicatedLine = [30, 30, 33, 28, 34, 29, 33, 28, 34, 29, 28, 33, 32]
# actualLine = [30, 30, 29, 33, 32, 29, 29, 33, 32, 29, 30, 31, 37]
# line_chart(xAxisvalue, predicatedLine, actualLine, title=title, theme='white')


def table_count(xAxisvalue, predicatedLine, actualLine,**optional_params):
    fig, ax = plt.subplots(figsize=(16, 9))

    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig,ax,theme=theme)

    x = np.arange(len(xAxisvalue))
    x_smooth = np.linspace(0, len(xAxisvalue) - 1, 100)
    f_actual = interp1d(x, actualLine, kind='cubic')
    f_predicted = interp1d(x, predicatedLine, kind='cubic')
    actualLine_smooth = f_actual(x_smooth)
    predicatedLine_smooth = f_predicted(x_smooth)
    ax.grid(True, linestyle='--', alpha=0.5) 
    ax.plot(x_smooth, actualLine_smooth, color="#FF8080", label='Actual')  
    ax.plot(x_smooth, predicatedLine_smooth, color="#80A0FF", label='Predicted')  
    legend = ax.legend(bbox_to_anchor=(1.02, 1), facecolor='black' if theme=='black' else 'white')  
    ax.xaxis.set_major_locator(plt.IndexLocator(base=1, offset=0))

    if theme=='black':
        legend.get_texts()[0].set_color('white') 
        legend.get_texts()[1].set_color('white')
        ax.set_xticklabels(xAxisvalue, color='white')
    else:
        ax.set_xticklabels(xAxisvalue, color='black')

    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])   
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel']) 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'],color='white')

    name = str(random.random())
    plt.savefig(name + '.png', facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(name + '.png')
    print("Image saved:", name + '.png')
    im.show()
    # plt.show()

# xAxisvalue = ['8:00 PM', '9:00 PM', '10:00 PM', '12:00 PM', '1:00 AM', '3:00 AM', '3:00 AM', '4:00 AM', '5:00 AM','3:00 AM', '3:00 AM', '4:00 AM', '5:00 AM']
# predicatedLine = [30, 30, 33, 28, 34, 29, 33, 28, 34, 29, 28, 33, 32]
# actualLine = [30, 30, 29, 33, 32, 29, 29, 33, 32, 29, 30, 31, 37]
# table_count(xAxisvalue, predicatedLine, actualLine,theme='white')

def dot_plot(categories,data,**optional_params):
    fig, ax = plt.subplots()

    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig,ax,theme=theme)

    ax.set_aspect('equal')
    ax.scatter(np.arange(len(categories)), data, s=100, c='#8566F8', alpha=0.7)
    ax.set_xticks(np.arange(len(categories)))
    ax.set_xticklabels(categories)

    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])   
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel']) 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'],color='white')

    name = random.random()
    plt.savefig(str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png') 
    print("Image saved: ", str(name) + '.png')
    im.show()
    # plt.show()  

# categories = [1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16]
# values = ['Option 1','Option 2','Option 3','Option 4','Option 5','Option 6','Option 7','Option 8','Option 9','Option 10','Option 11','Option 12','Option 13','Option 14','Option 15','Option 16']
# dot_plot(categories,values,theme='black')

def dot_plot_grouped(categories, data, **optional_params):
    colors = ["#FD93B1", "#FC2964", "#8566F8", "#5C33F6", "#E0E0E0", "#F2F2F2","#FD93B1", "#FC2964", "#FEC9D8"]
    fig, ax = plt.subplots()
    ax.set_aspect('equal')

    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig,ax,theme=theme)

    grouped_categories = [categories[i:i+4] for i in range(0, len(categories), 4)]
    group_count = len(grouped_categories)
    index=0

    for i, group_cats in enumerate(grouped_categories):
        ax.scatter(group_cats, data[i*4:(i+1)*4], s=100, c=colors[index], alpha=0.7)
        index=index+1

    ax.set_xticks(np.arange(len(categories)))
    ax.set_xticklabels(categories)
    ax.set_yticks(np.arange(len(categories)))
    y_labels = ax.set_yticklabels(data)
    ax.legend(np.arange(group_count)+1,title='legend',loc='lower center', bbox_to_anchor=(0.5, -0.3), ncol = 4)

    for i, label in enumerate(y_labels):
        if i < 4:
            label.set_color("#FD93B1")
        elif i < 8:
            label.set_color("#FC2964")
        elif i < 12:
            label.set_color("#8566F8")
        else:
            label.set_color("#5C33F6")

    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])   
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel']) 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'],color='white')

    name = random.random()
    plt.savefig(str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png')
    print("Image saved: ", str(name) + '.png')
    im.show()
    plt.close()
    # plt.show()

# categories = [ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
# values = ['Option 0', 'Option 1', 'Option 2', 'Option 3', 'Option 4', 'Option 5', 'Option 6', 'Option 7',
#           'Option 8', 'Option 9', 'Option 10', 'Option 11', 'Option 12', 'Option 13', 'Option 14', 'Option 15']

# dot_plot_grouped(categories, values,theme='white')


def scatter_plot_with_dotted_line(data,**optional_params):
    fig, ax = plt.subplots() 
    shapes = ["o", "v", "^", ".", ",", "<", ">", "1", "2", "3", "4", "8", "s", "p"]
    color = ["#8566F8","#8566F8","#666666","#BDBDBD","#FD93B1","#FEC9D8","#393939"]

    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig,ax,theme=theme)
    index=0

    for y in data:
        res = random.sample(range(1, len(y)+1), len(y))
        ax.scatter(res, y,color=color[index],marker=shapes[index])
        index=index+1

    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])
        ax.xaxis.label.set_color('white') 
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'])
        ax.yaxis.label.set_color('white') 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'],color='white')

    # if ('handles' in optional_params):
    #     ax.legend(handles=optional_params['handles'],ncol = 1 ,loc='upper left', bbox_to_anchor= (0.0, 1.010))   
    ax.axhline(y = (len(data)*len(data[0]))/2, color = '#FC2964', linestyle = ':')
    name = random.random()
    plt.savefig(str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png') 
    print("Image saved: ", str(name) + '.png')
    im.show()
    # plt.show()

# data=[[2,4,6,8,10,3,6,9,12,15,5,8,11,13,17,7,14,17,16,19,21,23,18,19,31,25,27,24,26,30,32,41,33,25,35,41,]]
# title = 'Scatter plot with dotted line'
# xlabel='Title'
# ylabel='Title'
# handles=['Group 1']
# scatter_plot_with_dotted_line(data,title=title,xlabel=xlabel,ylabel=ylabel, theme='white')


def scatter_plot_graph(marker_shapes, x_values, series_data, colors, **optional_params):
    fig, ax = plt.subplots(figsize=(11, 9))
    np.random.seed(19680801)
    sizes = np.random.rand(len(x_values[0])) * 50 + 50

    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig,ax,theme=theme)
    for i, marker in enumerate(marker_shapes):
        x = x_values[i]
        y = series_data[i]
        color = colors[i % len(colors)]
        ax.scatter(x, y, s=sizes, alpha=0.5, marker=marker, color=color, label="Group {}".format(i+1))

    legend = ax.legend(bbox_to_anchor=(1, 0.9), facecolor='black' if theme == 'black' else 'white')

    if theme == 'black':
        for text in legend.get_texts():
            text.set_color('white')

    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'], fontdict={'fontsize': 12})   
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'],fontdict={'fontsize': 12}) 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'],fontdict={'fontsize': 15})

    name = str(random.random())
    plt.savefig(name + '.png', facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(name + '.png')
    print("Image saved:", name + '.png')
    im.show()
    # plt.show()

# marker_shapes = ['s', 'o', 'D']
# x_values = [np.arange(0.15, 50.0, 2.0), np.arange(0.15, 50.0, 2.0), np.arange(0.15, 50.0, 2.0)]
# series_data = [
#     x_values[0] ** 1.3 + np.random.rand(*x_values[0].shape) * 30.0,
#     x_values[1] ** 1.5 + np.random.rand(*x_values[1].shape) * 40.0,
#     x_values[2] ** 1.2 + np.random.rand(*x_values[2].shape) * 20.0
# ]
# colors = ["#FC2964", "#0095FF", "#5C33F6"]
# scatter_plot_graph(marker_shapes, x_values, series_data, colors, theme='white')

def scatter_plot_with_line(data,**optional_params):
    fig, ax = plt.subplots()
    
    if 'theme' in optional_params:
        theme=optional_params['theme']
    if theme == 'black':
        set_graph_dark(fig,ax,theme=theme)

    shapes = ["o", "v", "^", ".", ",", "<", ">", "1", "2", "3", "4", "8", "s", "p"]
    color = ["#8566F8","#8566F8","#666666","#BDBDBD","#FD93B1","#FEC9D8","#393939"]
    index=0

    for y in data:
        res = random.sample(range(1, len(y)+1), len(y))
        ax.scatter(res, y,color=color[index],marker=shapes[index])
        index=index+1

    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])   
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel']) 
    if ('title' in optional_params):
        ax.set_title(optional_params['title'],color='white')
        
    # ax.legend(handles=handles,ncol = 1 ,loc='upper left', bbox_to_anchor= (0.0, 1.010))
    x1=np.array([1,(len(data)*len(data[0]))-1])
    y1=np.array([1,(len(data)*len(data[0]))-1])
    ax.plot(x1,y1,label='line',color = '#FC2964')
    name = random.random()
    plt.savefig(str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png') 
    print("Image saved: ", str(name) + '.png')
    im.show()
    # plt.show()

# data=[[2,4,6,8,10,3,6,9,12,15,5,8,11,13,17,7,14,17,16,19,21,23,18,19,31,25,27,24,26,30,32,41,33,25,35,41,]]
# title = 'Scatter plot with line'
# xlabel='Title'
# ylabel='Title'
# handles=['Group 1']
# scatter_plot_with_line(data,title=title,xlabel=xlabel,ylabel=ylabel)

# def treeMap(data,colors,title):
#     labels = [item for sublist in data for item in sublist]
#     sizes = [1] * len(labels)
#     fig = plt.figure(figsize=(8, 6))
#     ax = fig.add_subplot(111)
#     plt.title(title)
#     squarify.plot(sizes=sizes, label=labels, color=colors, alpha=0.7, ax=ax)
#     plt.axis('off')
#     name = random.random()
#     plt.savefig(str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
#     im = Image.open(str(name) + '.png') 
#     print("Image saved: ", str(name) + '.png')
#     im.show()

# data = [["property", "property", "property"],
#         ["property", "property", "property"],
#         ["property", "property"],
#         ["property"]]
# colors = ['#5C33F6', '#FEC9D8', '#BDBDBD']
# title="Tree Map"
# treeMap(data,colors,title)

def forest_plot_graph(yAxisdata, xAxisdata, effect_sizes, color,confidence_intervals):
    plt.figure(figsize=(6, 4))
    for i in range(len(yAxisdata)):
        if i < len(confidence_intervals):
            plt.plot([confidence_intervals[i][0], confidence_intervals[i][1]], [i, i], color=color)
            center = (confidence_intervals[i][0] + confidence_intervals[i][1]) / 2
            plt.scatter(center, i, color=color, zorder=10)
    x_vals = np.full(len(effect_sizes), np.nan) 
    y_vals = np.arange(len(effect_sizes))
    plt.scatter(effect_sizes, y_vals[:len(effect_sizes)], color=color, zorder=10)
    plt.yticks(range(len(yAxisdata)), yAxisdata)
    plt.xticks(xAxisdata)
    plt.xlim(min(xAxisdata), max(xAxisdata))
    plt.title('Forest Plot')
    plt.xlabel('Title')
    plt.ylabel('Title')
    name = random.random()
    # plt.savefig(str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(str(name) + '.png') 
    print("Image saved: ", str(name) + '.png')
    im.show()

# yAxisdata = [10, 20, 30, 40, 50, 60, 70]
# xAxisdata = [10, 20, 30, 40, 50, 60, 70, 80]
# effect_sizes = [13, 20, 30, 40, 50]
# confidence_intervals = [(10, 20), (20, 30), (30, 40), (40, 50), (50, 60)]
# color="#5C33F6"
# forest_plot_graph(yAxisdata, xAxisdata, effect_sizes,color, confidence_intervals)


# def box_plot(data,**optional_params):
#     fig, ax = plt.subplots() 
#     colors = ['#8566F8','#E6F5FF']
#     length_of_data = len(data)

#     boxplot = ax.boxplot(data, vert=True, patch_artist=True)
#     ax.grid(color=colors[1])
#     boxplot['whiskers'][0].set(color=colors[0])
#     boxplot['whiskers'][1].set(color=colors[0])
#     boxplot['caps'][0].set(color=colors[0])
#     boxplot['caps'][1].set(color=colors[0])
#     ax.set_xticklabels(np.arange(length_of_data))

#     for box in boxplot['boxes']:
#         box.set_facecolor(colors[0])
#         box.set_edgecolor(colors[0])

#     # for whisker in boxplot['whiskers']:
#     #     whisker.set_linestyle('') 

#     if('y_label' in optional_params):
#         ax.set_ylabel(optional_params['y_label'])
#     if('x_label' in optional_params):
#         ax.set_xlabel(optional_params['x_label'])
#     if('title' in optional_params):
#         ax.set_title(optional_params['title'])
    
#     name = random.random()
#     plt.savefig(str(name) + '.png' , facecolor='black' if theme == 'black'else 'white' )
#     im = Image.open(str(name) + '.png') 
#     print("Image saved: ", str(name) + '.png')
#     im.show()
#     plt.show()

# data=[[25, 30],[20, 27],[22, 27],[25, 28],[19, 28],[25, 40],[30, 40],[38, 50],[40, 47],[40, 50],[38, 55]]
# title = 'Box plots'
# x_label='Title'
# y_label='Title'

# box_plot(data,title=title,x_label=x_label,y_label=y_label)

def insertion_sort(header, header_data):
    length = len(header_data)
    indices = list(range(length))
    indices.sort(key=lambda i: abs(header_data[i]), reverse=True)
    final_headers = []
    final_headers_data = []
    my_slice = slice(0,  9 if len(header)> 9 else len(header))
    indexes = indices[my_slice]
    for index in indexes:
        final_headers.append(header[index])
        final_headers_data.append(header_data[index])
    
    header = final_headers
    header_data = final_headers_data
    
    for i in range(len(header_data)):
        current = header_data[i]
        current_header = header[i]
        j = i - 1
        
        while j >= 0 and header_data[j] > current:
            header_data[j + 1] = header_data[j]
            header[j + 1] = header[j]
            j -= 1
        
        header_data[j + 1] = current
        header[j + 1] = current_header
    
    # header_data.reverse()
    # header.reverse()
    
    return {'categories':header, 'values':  header_data}


def explanability_chart(data, **optional_params):
    flattened_list =[]
    for sublist in  list(data["inputs"].values()):
        for item in sublist:
            flattened_list.append(item)
    answer =  insertion_sort(list(data['inputs'].keys()), flattened_list)
    fig, ax = plt.subplots()
    print(answer)
    theme ='black'
    if 'theme' in optional_params:
        theme= optional_params['theme']
    categories = [c for c, v in zip(answer['categories'],answer['values']) if v is not None]
    values = [v for v in answer['values'] if v is not None]
    print(values)
    if 'title' in optional_params:
        ax.set_title(optional_params['title'], fontdict={'fontsize': 15, 'color': ('black' if theme != 'black'else 'white')})
    bars = ax.barh(categories, values, align='center')

    ax.set_xlabel('Value', fontdict={'fontsize': 12, 'color':  ('black' if theme != 'black'else 'white')})
    ax.set_ylabel('Categories', fontdict={'fontsize': 8, 'color':  ('black' if theme != 'black'else 'white')})
    ax.set_title('FEATURE CONTRIBUTION', fontdict={'fontsize': 15, 'color':  ('black' if theme != 'black'else 'white')})
    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'])
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])
    if ('legend' in optional_params):
        ax.legend(title=optional_params['legend'])
    if theme == 'black':
        ax.spines['top'].set_visible(True) 
        ax.spines['right'].set_visible(True) 
        ax.spines['bottom'].set_visible(True) 
        ax.spines['left'].set_visible(True) 

        ax.spines['top'].set_color('white') 
        ax.spines['right'].set_color('white') 
        ax.spines['bottom'].set_color('white')  
        ax.spines['left'].set_color('white') 

        ax.spines['top'].set_linewidth(0.8)  
        ax.spines['right'].set_linewidth(0.8)  
        ax.spines['bottom'].set_linewidth(0.8) 
        ax.spines['left'].set_linewidth(0.8)  

    ax.set_axisbelow(True)
    color=('black' if theme != 'black'else 'white')
    ax.grid(axis='x', color=color  , linestyle='--', alpha=0.5)
    for bar, value in zip(bars, answer['values']):
        if value < 0:
            bar.set_color('#FC2964')
        else:
            bar.set_color('#0095FF')

    ax.set_yticklabels(categories, fontdict={'fontsize': 8, 'color': color })
    ax.set_xticklabels(ax.get_xticks(), fontdict={'fontsize': 8, 'color': color })
    print(data['outputs'][0])
    ax.set_ylim( data['outputs'][0])
    ax.set_facecolor('black' if theme == 'black'else 'white')
    fig.patch.set_facecolor('black' if theme == 'black'else 'white')
    if ('xlabel' in optional_params):
        ax.set_xlabel(optional_params['xlabel'])
    if ('ylabel' in optional_params):
        ax.set_ylabel(optional_params['ylabel'])
    if ('title' in optional_params):
        ax.set_title(optional_params['title'])
    if ('legend' in optional_params):
        ax.legend(title=optional_params['legend'])
    name = str(random.random())
    plt.savefig(name + '.png',  bbox_inches='tight',facecolor='black' if theme == 'black'else 'white' )
    im = Image.open(name + '.png')
    print("Image saved:", name + '.png')
    im.show()




# categories = ['Burger', 'Pizza', 'sandwich', 'pasta', 'momos', 'garlic']
# values = [-10, -5, -4, 4, 5, 10]
# title = 'Secondary text'
data = {
"inputs": { "age": [0.051], "capital-loss": [-0.006], "education": [-0.057], "education-num": [-0.034], "hours-per-week": [0.174], "native-country": [0.006], "occupation": [-0.03], "relationship": [0.163], "workclass": [0] }, "outputs": [0.267]
}
explanability_chart(data)




import pandas as pd, numpy as np
from scipy.stats import tstd

def main(d_params, dataset):
    """ FORMAT THE DATA & SAVE VARIABLES FOR LATER REFERENCE

    Args: 
        d_params: data parameters, see train_model() above
        dataset: pd.DataFrame - table of all of the samples (training, validation, & testing)

    Returns:
        x_train: dictionary - see above load_dataset()"""

    # Cannot Train W/O Labels

    dataset = dataset.dropna(subset=list(d_params["y"])).reset_index(drop=True)

    data_mapping = {}
    for k, val in {**d_params["x"], **d_params["y"]}.items():

        if val["type"] == "NUMBER":  # NUMERICAL

            dataset[k] = pd.to_numeric(dataset[k], errors='coerce').astype(float) # change dtype

            # Create a Histogram
            tmp = dataset[k].copy().dropna()  # cant use nan for number histograms
            minim, maxim = min(tmp), max(tmp)  # max-min does not persist
            if not minim == maxim:
                bins = int(round((maxim - minim) / (3.49 * tstd(tmp) * (len(tmp) ** -.333333))))
                tmp = np.histogram(tmp, bins, range=(minim, maxim))
                hist_a = list(tmp[0]/tmp[0].sum())
                data_mapping = {**data_mapping, k: {"min": minim, "max": maxim, "hist": hist_a, "type": val["type"], "bins": bins}}
            else:
                data_mapping = {**data_mapping, k: {"min": minim, "max": maxim, "hist": [maxim], "type": val["type"], "bins": 1}}

            # NANs FOR NUMERICAL
            if "null" in list(val):
                if val["null"] == "mean":
                    dataset[k] = dataset[k].fillna(dataset[k].mean())  # assume mean value
                else:
                    raise Exception(f"Error: {val['null']} not valid for {k}")
            else:
                dataset = dataset[dataset[k].notna()]  # drop nans

        elif val["type"] == "CATEGORY":  # CATEGORICAL, data shall be enumerated
            
            dataset[k] = dataset[k].astype(str)  # in case inconsistent input types with same value (e.g. 9 and "9")

            # Create analytics from non-nan
            if k in d_params["x"].keys():  # for inputs
                categories = ['nan'] + [i for i in sorted(list(set(dataset[k]))) if i != 'nan']
                dataset[k] = dataset[k].replace({val: k for k, val in enumerate(categories)}).astype('int')
                tmp = dataset[k][dataset[k] != 0].copy() # dont use nan b/c ruin analytics for sparse
            elif k in d_params["y"].keys():  # for outputs
                categories = [i for i in sorted(list(set(dataset[k]))) if i != 'nan']
                dataset[k] = dataset[k].replace({val: k for k, val in enumerate(categories)}).astype('int')
                tmp = dataset[k].copy()
            else:
                raise Exception(f"Error: problem with column {k} on learn.py data_validation()")

            # Histogram
            tmp = np.histogram(tmp, bins=len(categories)) # disclude nan cat.
            hist_a = list(tmp[0]/tmp[0].sum())
            data_mapping = {**data_mapping, k: {"hist": hist_a, "cats": categories, "type": val["type"]}}


        elif val["type"] == "STRING":  # STRING (cannot create histogram from non-categorical text)

            dataset[k] = dataset[k].astype(str).replace({"": 'nan'})  # TODO: is this right order?
            data_mapping = {**data_mapping, k: {"type": val["type"]}}

        else:
            raise Exception(f"dtype '{val['type']}' is not valid")

    dataset.reset_index(inplace=True, drop=True)

    return dataset[list(d_params["x"])], dataset[list(d_params["y"])], data_mapping



import pandas as pd
from catboost import CatBoostRegressor
from lightgbm import LGBMRegressor
from sklearn.ensemble import (ExtraTreesRegressor, GradientBoostingRegressor, RandomForestRegressor)
from sklearn.linear_model import ElasticNet
from sklearn.tree import DecisionTreeRegressor
from xgboost.sklearn import XGBRegressor


# def get_dict_regress_models(seed=42):
#     '''returns a dictionary of models to be used in the training process'''

#     return {
#         'CATBOOST': CatBoostRegressor(verbose=False, random_state=seed),
#         'ELAST': ElasticNet(random_state=seed),
#         'GRADB': GradientBoostingRegressor(random_state=seed),
#         'RFR': RandomForestRegressor(verbose=1, random_state=seed),
#         'DTR': DecisionTreeRegressor(random_state=seed),
#         'XTR': ExtraTreesRegressor(random_state=seed),
#         'LGBM': LGBMRegressor(random_state=seed),
#         'XGB': XGBRegressor(use_label_encoder=False, eval_metric='error', random_state=seed),
#     }


# def create_model(reg_name, **kwargs):
#     '''returns an instantiated model to be used in the training process'''
#     try:
#         reg = get_dict_regress_models()[reg_name]
#         reg.set_params(**kwargs)
#         return reg
#     except Exception as e:
#         logger.error(f"> Error create_model({reg_name}): {e}")
#     return None


# def train_the_model(reg, x, y, use_ravel=True):
#     '''Train the model using the X and Y values. If use_ravel is True, the Y values will be raveled'''
#     try:
#         reg.fit(x, y.ravel() if use_ravel else y)
#         return reg
#     except Exception as e:
#         logger.error(f"> Error train_the_model({reg}): {e}")
#     return None


# def just_predict_model(pipe, x, scaler_y=None):
#     '''Predict X values using the model. If scaler_y is not None, it will be used to inverse transform the predicted values'''
#     predicted = pipe.predict(x)
#     if scaler_y is not None:
#         predicted = scaler_y.inverse_transform(predicted.reshape(-1, 1))
#     return predicted


# def predict_model(pipe, x_train, x_test, scaler_y=None):
#     '''Predict train and test dataset using the model. If scaler_y is not None, it will be used to inverse transform the predicted values'''
#     predicted = pipe.predict(x_train)
#     predicted_t = pipe.predict(x_test)
#     if scaler_y is not None:
#         predicted = scaler_y.inverse_transform(predicted.reshape(-1, 1))
#         predicted_t = scaler_y.inverse_transform(predicted_t.reshape(-1, 1))
#     return predicted, predicted_t


# def calc_mape_mae(truth, pred):
#     '''Calculate MAPE and MAE from truth and pred values'''
#     df = pd.DataFrame({'truth': truth.ravel(), 'pred': pred.ravel()})
#     # Replace negative predictions with 0
#     df['pred'] = df['pred'].clip(lower=0)
#     df['error'] = abs(df['pred'] - df['truth'])
#     mape = df['error'].sum() / df['truth'].sum()
#     mae = df['error'].mean()
#     return mape, mae, df



import pandas as pd

def dataframe(d_params, credentials, s3=None):
    """LOAD THE CSVS FROM AN S3 BUCKET

    Args:
        d_params: data parameters, see train_model() above
        s3: s3 connection
        credentials: dictionary - confidential information used to access s3 buckets 

    Returns:
        xy: pd.DataFrame - table of input features + outputs
        
    """

    # Download, Concat, & Reset Indices
    datasets = []

    if s3:
        for i, ds in enumerate(d_params["file_paths"]):  # csv's stored by date

            date = ds.split("/")[-1].split(".csv")[0]  
            s3.download_file(credentials["bucket"], ds, f"/tmp/{date+'.csv'}")
            dataset = pd.read_csv(f"/tmp/{date+'.csv'}", index_col=False)
            
            # if sorting datetime, need datetime
            if "timeseries" in d_params and list(d_params["timeseries"]["sort"])[0] == "datetime":
                dataset["datetime"] = pd.to_datetime(dataset["HH:MM:SS"]+" "+date)
                dataset.set_index("datetime", inplace=True, drop=True)
            
            datasets += [dataset] 

    # Join datasets
    columns = list({**d_params["x"], **d_params["y"]})
    datasets = pd.concat(datasets)[columns]

    return datasets
    



from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler, MinMaxScaler
# from func.helpers import DEP
from func.preprocess import check_ds_config


# def train_test_sp(df):
#     # Split the dataset into train and test
#     pos_to_split = int(len(df) * 0.8)
#     df = df.reset_index(drop=False)
#     df_train = df[df.index < pos_to_split].set_index('cd')
#     df_test = df[df.index >= pos_to_split].set_index('cd')

#     # Split the dataset into x_train x_test, and y_train y_test
#     x_train = df_train.drop(columns=[DEP]).copy()
#     y_train = df_train[DEP].copy()
#     x_test = df_test.drop(columns=[DEP]).copy()
#     y_test = df_test[DEP].copy()
#     return x_train, x_test, y_train, y_test


# def scale_and_pca_df(x, y, scaler_x=None, scaler_y=None, scaler_pca=None, conf=None):
#     '''Scale and PCA the dataset. 
#     If scaler_x, scaler_y or scaler_pca are  None: We do fit_transform
#     If scaler_x, scaler_y or scaler_pca are not None: We just do transform
#     Returns x and y (scaled and reduced), scaler_x, scaler_y, scaler_pca
#     '''
#     conf = check_ds_config(conf)
#     y = y.to_numpy().reshape(-1, 1)
#     # Scaling the dataset
#     if conf['do_scale']:
#         if scaler_x is None:
#             scaler_x = StandardScaler(
#             ) if conf['scaler_x'] == 'std' else MinMaxScaler()
#             x = scaler_x.fit_transform(x)
#         else:
#             x = scaler_x.transform(x)

#     if conf['scale_y']:
#         if scaler_y is None:
#             scaler_y = StandardScaler(
#             ) if conf['scaler_y'] == 'std' else MinMaxScaler()
#             y = scaler_y.fit_transform(y)
#         else:
#             y = scaler_y.transform(y)

#     if conf['do_pca']:
#         if scaler_pca is None:
#             scaler_pca = PCA(n_components=.995)
#             x = scaler_pca.fit_transform(x)
#         else:
#             x = scaler_pca.transform(x)

#     return x, y, scaler_x, scaler_y, scaler_pca
